/*
 * D-Bus interface
 *
 * Copyright (C) 2005 TransGaming Technologies Inc
 */
#include "config.h"

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#include <unistd.h>
#include <fcntl.h>
#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif

#include "config.h"
#include "wine/port.h"
#include "object.h"
#include "thread.h"
#include "server_memory.h"
#include "request.h"
#include "process.h"
#include "handle.h"
#include "mapping.h"
#include "file.h"
#include "cdrom_eject.h"

/* Until version 1.0, D-Bus API is subject to change */
/* #define DBUS_API_SUBJECT_TO_CHANGE */
/* #include <dbus/dbus.h> */

/** D-Bus Headers from dbus 0.23.4
 */

typedef unsigned int   dbus_bool_t;
typedef unsigned short dbus_uint16_t;
typedef short          dbus_int16_t;
typedef unsigned int   dbus_uint32_t;
typedef int            dbus_int32_t;
typedef dbus_uint32_t  dbus_unichar_t;

typedef struct DBusWatch DBusWatch;
typedef struct DBusConnection DBusConnection;
typedef struct DBusError DBusError;
typedef struct DBusTimeout DBusTimeout;
typedef struct DBusMessage DBusMessage;
typedef struct DBusMessageIter DBusMessageIter;

typedef enum
{
  DBUS_BUS_SESSION,    /**< The login session bus */
  DBUS_BUS_SYSTEM,     /**< The systemwide bus */
  DBUS_BUS_ACTIVATION  /**< The bus that activated us, if any */
} DBusBusType;

typedef enum
{
  DBUS_HANDLER_RESULT_HANDLED,         /**< Message has had its effect */ 
  DBUS_HANDLER_RESULT_NOT_YET_HANDLED, /**< Message has not had any effect */
  DBUS_HANDLER_RESULT_NEED_MEMORY      /**< Need more memory to return another result */
} DBusHandlerResult;

typedef enum
{
  DBUS_DISPATCH_DATA_REMAINS,  /**< There is more data to potentially convert to messages. */
  DBUS_DISPATCH_COMPLETE,      /**< All currently available data has been processed. */
  DBUS_DISPATCH_NEED_MEMORY    /**< More memory is needed to continue. */
} DBusDispatchStatus;

typedef enum
{
  DBUS_WATCH_READABLE = 1 << 0, /**< As in POLLIN */
  DBUS_WATCH_WRITABLE = 1 << 1, /**< As in POLLOUT */
  DBUS_WATCH_ERROR    = 1 << 2, /**< As in POLLERR (can't watch for this, but
                                 *   the flag can be passed to dbus_watch_handle()).
                                 */
  DBUS_WATCH_HANGUP   = 1 << 3  /**< As in POLLHUP (can't watch for it, but
                                 *   can be present in current state). */
} DBusWatchFlags;

typedef dbus_bool_t (* DBusAddWatchFunction)       (DBusWatch      *watch,
                                                    void           *data);
typedef void        (* DBusWatchToggledFunction)   (DBusWatch      *watch,
                                                    void           *data);
typedef void        (* DBusRemoveWatchFunction)    (DBusWatch      *watch,
                                                    void           *data);
typedef dbus_bool_t (* DBusAddTimeoutFunction)     (DBusTimeout    *timeout,
                                                    void           *data);
typedef void        (* DBusTimeoutToggledFunction) (DBusTimeout    *timeout,
                                                    void           *data);
typedef void        (* DBusRemoveTimeoutFunction)  (DBusTimeout    *timeout,
                                                    void           *data);
typedef void        (* DBusDispatchStatusFunction) (DBusConnection *connection,
                                                    DBusDispatchStatus new_status,
                                                    void           *data);
typedef void        (* DBusWakeupMainFunction)     (void           *data);
typedef DBusHandlerResult (* DBusHandleMessageFunction) (DBusConnection     *connection,
                                                         DBusMessage        *message,
                                                         void               *user_data);
typedef void (* DBusFreeFunction) (void *memory);

#define DBUS_TYPE_INVALID       ((int) '\0')
#define DBUS_TYPE_STRING        ((int) 's')

#define DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL "org.freedesktop.Local"

struct DBusError
{
  const char *name;    /**< error name */
  const char *message; /**< error message */

  unsigned int dummy1 : 1; /**< placeholder */
  unsigned int dummy2 : 1; /**< placeholder */
  unsigned int dummy3 : 1; /**< placeholder */
  unsigned int dummy4 : 1; /**< placeholder */
  unsigned int dummy5 : 1; /**< placeholder */

  void *padding1; /**< placeholder */
};

/**
 * DBusMessageIter struct; contains no public fields 
 */
struct DBusMessageIter
{ 
  void *dummy1;         /**< Don't use this */
  void *dummy2;         /**< Don't use this */
  dbus_uint32_t dummy3; /**< Don't use this */
  int dummy4;           /**< Don't use this */
  int dummy5;           /**< Don't use this */
  int dummy6;           /**< Don't use this */
  int dummy7;           /**< Don't use this */
  int dummy8;           /**< Don't use this */
  int dummy9;           /**< Don't use this */
  int dummy10;          /**< Don't use this */
  int dummy11;          /**< Don't use this */
  int pad1;             /**< Don't use this */
  int pad2;             /**< Don't use this */
  void *pad3;           /**< Don't use this */
};


/** Implementation structures
 */

struct dbus_lib_table
{
  void (* dbus_error_init) (DBusError *error);
  void (* dbus_error_free) (DBusError *error);
  dbus_bool_t (* dbus_error_is_set) (const DBusError *error);

  DBusConnection * (* dbus_bus_get) (DBusBusType type, DBusError *error);
  void (* dbus_bus_add_match) (DBusConnection *connection, const char *rule, DBusError *error);

  DBusConnection * (* dbus_connection_ref) (DBusConnection *connection);
  void             (* dbus_connection_unref) (DBusConnection *connection);
  DBusDispatchStatus (* dbus_connection_dispatch) (DBusConnection *connection);
  dbus_bool_t (* dbus_connection_set_watch_functions) (DBusConnection *connection, 
      DBusAddWatchFunction add_function, DBusRemoveWatchFunction remove_function, 
      DBusWatchToggledFunction toggled_function, void *data, 
      DBusFreeFunction free_data_function);
  /*
  dbus_bool_t (* dbus_connection_set_timeout_functions) (DBusConnection *connection, 
      DBusAddTimeoutFunction add_function, DBusRemoveTimeoutFunction remove_function, 
      DBusTimeoutToggledFunction toggled_function, void *data, 
      DBusFreeFunction free_data_function);
  void (* dbus_connection_set_wakeup_main_function) (DBusConnection *connection, 
      DBusWakeupMainFunction wakeup_main_function, void *data, 
      DBusFreeFunction free_data_function);
  */
  dbus_bool_t (* dbus_connection_add_filter) (DBusConnection *connection, 
      DBusHandleMessageFunction function, void *user_data, 
      DBusFreeFunction free_data_function);
  void        (*dbus_connection_remove_filter) (DBusConnection *connection,
      DBusHandleMessageFunction function, void *user_data);
  void (* dbus_connection_disconnect) (DBusConnection *connection);

  const char * (* dbus_message_get_path) (DBusMessage *message);
  const char * (* dbus_message_get_interface) (DBusMessage *message);
  const char * (* dbus_message_get_member) (DBusMessage *message);
  dbus_bool_t  (* dbus_message_is_signal) (DBusMessage *message, const char *interface, const char *signal_name);
  dbus_bool_t  (* dbus_message_get_args) (DBusMessage *message, DBusError *error, int first_arg_type,...);
  dbus_bool_t  (* dbus_message_iter_init) (DBusMessage *message, DBusMessageIter *iter);
  void         (* dbus_message_iter_get_basic) (DBusMessageIter *iter, void *value);
  dbus_bool_t  (* dbus_message_iter_has_next) (DBusMessageIter *iter);
  dbus_bool_t  (* dbus_message_iter_next) (DBusMessageIter *iter);
  void         (* dbus_message_iter_recurse) (DBusMessageIter *iter, DBusMessageIter *sub);

  int          (* dbus_watch_get_fd) (DBusWatch *watch);
  unsigned int (* dbus_watch_get_flags) (DBusWatch *watch);
  dbus_bool_t  (* dbus_watch_get_enabled) (DBusWatch *watch);
  dbus_bool_t  (* dbus_watch_handle) (DBusWatch *watch, unsigned int flags);
  void* (* dbus_watch_get_data) (DBusWatch *watch);
  void  (* dbus_watch_set_data) (DBusWatch *watch, void *data, DBusFreeFunction free_data_function);
};

/* this table wraps a pollfd created for dbus ... */
struct dbus_table
{
  struct object    obj;
  int              fd;
  short            flags;
  dbus_int32_t     on;
  dbus_int32_t     was_on;
  DBusWatch        *watch;
};


/** Global variables
 */

static void dbus_table_dump( struct object *obj, int verbose );
static void dbus_poll_event(struct object *obj, int event);
/* static int  dbus_get_fd(struct object *obj); */
static void dbus_destroy( struct object *obj );

static const struct object_ops dbus_table_ops =
{
  sizeof(struct dbus_table),       /* size */
  dbus_table_dump,                 /* dump */
  no_add_queue,                    /* add_queue */
  NULL,                            /* remove_queue */
  NULL,                            /* signaled */
  no_signal,                       /* signal */
  NULL,                            /* satisfied */
  NULL,                            /* get_poll_events */
  dbus_poll_event,                 /* poll_event */
  no_get_fd,                     /* get_fd */
  no_get_ctx_fd,                   /* get_ctx_fd */
  no_flush,                        /* flush */
  no_get_file_info,                /* get_file_info */
  NULL,                            /* queue_async */
  dbus_destroy                     /* destroy */
};
const struct object_ops* get_this_process_dbus_table_ops(void) { return &dbus_table_ops; }

START_FILE_STATIC_VARIABLES(dbus)
 void                    *dbus_lib;
 struct dbus_lib_table    dbus_ltable;
 DBusError                dbus_error;
 DBusConnection          *dbus_conn;
 struct dbus_table       *dbus_list;
 int                      dbus_version;
END_FILE_STATIC_VARIABLES(dbus)

#define DBUS_VERSION_NONE    0
#define DBUS_VERSION_0_23_4  1
#define DBUS_VERSION_0_35_2  2
#define DBUS_VERSION_0_60_0  3
#define DBUS_VERSION_1_0_0   4

/**
 *  Code!
 **/

/* This loads the vtable for the dbus lib */
void load_dbus_table(void)
{
#define dset(x)  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).x = \
                    wine_dlsym(SERVER_STATIC_VARIABLE(dbus, dbus_lib), #x, NULL, 0)
  dset(dbus_error_init);
  dset(dbus_error_free);
  dset(dbus_error_is_set);

  dset(dbus_bus_get);
  dset(dbus_bus_add_match);

  dset(dbus_connection_ref);
  dset(dbus_connection_unref);
  dset(dbus_connection_dispatch);
  dset(dbus_connection_set_watch_functions);
  /* dset(dbus_connection_set_timeout_functions); */
  /* dset(dbus_connection_set_wakeup_main_function); */
  dset(dbus_connection_add_filter);
  dset(dbus_connection_remove_filter);
  dset(dbus_connection_disconnect);

  dset(dbus_message_get_path);
  dset(dbus_message_get_interface);
  dset(dbus_message_get_member);
  dset(dbus_message_is_signal);
  dset(dbus_message_get_args);
  dset(dbus_message_iter_init);
  dset(dbus_message_iter_get_basic); /* D-Bus 0.3x + only */
  dset(dbus_message_iter_has_next);
  dset(dbus_message_iter_next);
  dset(dbus_message_iter_recurse); /* D-Bus 0.3x + only */

  dset(dbus_watch_get_fd);
  dset(dbus_watch_get_flags);
  dset(dbus_watch_get_enabled);
  dset(dbus_watch_handle);
  dset(dbus_watch_get_data);
  dset(dbus_watch_set_data);
#undef dset
}


/** Handler for watching for a changed mount point.
 *  For HAL 0.4.7 we need to watch for:
 *   signal interface=org.freedesktop.Hal.Device; member=Condition;
 *   string:VolumeMount
 *   string:/dev/hdc
 *   string:/media/BG2_CD1
 *  For HAL 0.5.4 we need to watch for:
 *   signal interface=org.freedesktop.Hal.Device; member=PropertyModified;
 *  For HAL 0.60 we need to also watch for eject notifications:
 *   signal sender=:1.1 -> dest=(null destination) interface=org.freedesktop.Hal.Device; 
 *          member=Condition string "EjectPressed"
 */
static DBusHandlerResult
hal_filter_func (DBusConnection     *connection,
                 DBusMessage        *message,
                 void               *user_data)
{
  /*
  TRACE("FILTER path='%s' i='%s' mem='%s'\n",
         dbus.dbus_message_get_path(message),
         dbus.dbus_message_get_interface(message),
         dbus.dbus_message_get_member(message));
  */

  /* HAL 0.4.7 mount/unmount is checked this way */
  if ((SERVER_STATIC_VARIABLE(dbus, dbus_version) == DBUS_VERSION_0_23_4) &&
      SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_is_signal(
        message, 
        "org.freedesktop.Hal.Device",
        "Condition"))
  {
    const char *mnt_status, *mnt_device, *mnt_path;
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_get_args( 
        message, &SERVER_STATIC_VARIABLE(dbus, dbus_error),
        DBUS_TYPE_STRING, &mnt_status,
        DBUS_TYPE_STRING, &mnt_device,
        DBUS_TYPE_STRING, &mnt_path,
        DBUS_TYPE_INVALID );

    /* if we mounted or unmounted, take appropriate action */
    if(strcmp("VolumeMount", mnt_status) == 0) {
      if (debug_level) fprintf(stderr, "Drive has been mounted\n");
      cdrom_eject_dbus(mnt_device, TRUE);
    } else if(strcmp("VolumeUnmount", mnt_status) == 0) {
      if (debug_level) fprintf(stderr, "Drive has been unmounted\n");
      /* cdrom_eject_dbus(mnt_device, FALSE); */
    }

    return DBUS_HANDLER_RESULT_HANDLED;
  }
  
  /* HAL 0.5.4 mount/unmount doesn't send the other message, so.. */
  if ((SERVER_STATIC_VARIABLE(dbus, dbus_version) == DBUS_VERSION_0_35_2 ||
       SERVER_STATIC_VARIABLE(dbus, dbus_version) >= DBUS_VERSION_0_60_0) &&
      SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_is_signal(
      message,
      "org.freedesktop.Hal.Device",
      "PropertyModified"))
  {
    DBusMessageIter iter;
    DBusMessageIter iter_array;
    DBusMessageIter iter_struct;
    int num_items, i;
    int has_next;
    char *key;

    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_init (message, &iter);
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_get_basic (&iter, &num_items);
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_next (&iter);

    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_recurse (&iter, &iter_array);
    has_next = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_has_next(&iter_array);

    for (i=0; i<num_items; i++)
    {
      SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_recurse (&iter_array, &iter_struct);
      SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_get_basic (&iter_struct, &key);

      if (strcmp(key, "volume.mount_point")==0)
      {
        if (debug_level) fprintf(stderr, "Drive mountpoint has changed\n");
        cdrom_eject_dbus(NULL, TRUE);
        return DBUS_HANDLER_RESULT_HANDLED;
      }

      has_next = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_iter_next (&iter_array);
    }
  }

  if ((SERVER_STATIC_VARIABLE(dbus, dbus_version) >= DBUS_VERSION_0_60_0) &&
      SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_is_signal(
      message,
      "org.freedesktop.Hal.Device",
      "Condition"))
  {
    const char *condition_string;
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_get_args( 
        message, &SERVER_STATIC_VARIABLE(dbus, dbus_error),
        DBUS_TYPE_STRING, &condition_string,
        DBUS_TYPE_INVALID );

    /* if we mounted or unmounted, take appropriate action */
    if(strcmp("EjectPressed", condition_string) == 0) {
      if (debug_level) fprintf(stderr, "Drive has been ejected\n");
      cdrom_eject_dbus("", FALSE);
    }
  }

  if (SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_message_is_signal(
      message,
      DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL,
      "Disconnected"))
    if (debug_level) fprintf(stderr, "Disconnected from dbus.\n");

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}


/**
 *  Callbacks from D-Bus lib
 **/
dbus_uint32_t add_dbus_watch(DBusWatch *watch, void *data)
{
  dbus_uint32_t  cond   =  POLLHUP | POLLERR;
  dbus_int32_t flags;
  struct dbus_table *n;
  struct dbus_table **ptr;

  if (debug_level) fprintf(stderr,"dbus add\n");
  flags = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_get_flags(watch);

  if (flags & DBUS_WATCH_READABLE) 
    cond |= POLLIN;
  if (flags & DBUS_WATCH_WRITABLE) 
    cond |= POLLOUT;

  /* create a new object, add to linked list */
  n = alloc_object(dbus_table_ops_tag, SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_get_fd(watch));
  if (!n)
    return 0;
  SERVER_STATIC_VARIABLE(dbus, dbus_list) = n;
  n->fd = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_get_fd(watch);
  n->flags = cond;
  n->watch = watch;
  n->on = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_get_enabled(watch);

  /* add pointer to object as watch data */
  ptr = malloc(sizeof(struct dbus_table **));
  *ptr = n;
  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_set_data(watch, ptr, (DBusFreeFunction)free);

  /* add object to select */
  set_select_events(&n->obj, n->on ? n->flags : 0, NULL);

  return 1;
}

void remove_dbus_watch(DBusWatch *watch, void *data)
{
  struct dbus_table *ptr;

  if (debug_level) fprintf(stderr,"dbus remove\n");

  ptr = *((struct dbus_table **)(SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_get_data(watch)));

  /* deallocate object */
  release_object(ptr);
}

void watch_dbus_toggled(DBusWatch *watch, void *data)
{
  struct dbus_table *ptr;
  ptr = *((struct dbus_table **)data);

  if (debug_level) fprintf(stderr, "dbus toggled\n");

  ptr->on = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_get_enabled(watch);
  set_select_events(&ptr->obj, ptr->on ? ptr->flags : 0, NULL);
}


/**
 *  Callbacks from wineserver
 **/
static void dbus_poll_event(struct object *obj, int event)
{
  struct dbus_table *ptr = (struct dbus_table *)obj;
  dbus_uint32_t flags   = 0;

  assert( obj->ops == dbus_table_ops_tag );

  if (event & POLLIN)  
    flags |= DBUS_WATCH_READABLE;
  if (event & POLLOUT) 
    flags |= DBUS_WATCH_WRITABLE;
  if (event & POLLHUP) 
    flags |= DBUS_WATCH_HANGUP;
  if (event & POLLERR) 
    flags |= DBUS_WATCH_ERROR;

  if (!flags) return;

  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_ref(SERVER_STATIC_VARIABLE(dbus, dbus_conn));

  /* Do whatever we were waiting on the watch to do */
  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_watch_handle(ptr->watch, flags);

  /* Dispatch messages */
  while (SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_dispatch(
         SERVER_STATIC_VARIABLE(dbus, dbus_conn)) == DBUS_DISPATCH_DATA_REMAINS);

  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_unref(SERVER_STATIC_VARIABLE(dbus, dbus_conn));
}

static void dbus_table_dump( struct object *obj, int verbose )
{
  struct dbus_table *table = (struct dbus_table *)obj;

  assert( obj->ops == dbus_table_ops_tag );

  if (debug_level) fprintf( stderr, "D-Bus watch table fd=%d flags=%d on=%s\n",
                            table->fd, table->flags, table->on ? "yes":"no" );
}

#if 0
static int  dbus_get_fd(struct object *obj)
{
  struct dbus_table *table = (struct dbus_table *)obj;

  assert( obj->ops == dbus_table_ops_tag );

  return table->fd;
}
#endif

static void dbus_destroy( struct object *obj )
{
  struct dbus_table *table = (struct dbus_table *)obj;

  assert( obj->ops == dbus_table_ops_tag );

  if (debug_level) fprintf(stderr, "dbus destroy\n");
  table->watch = NULL;
}




void dbus_init(void)
{
  if (debug_level) fprintf(stderr, "starting dbus..\n");
  /* server variables */
  SERVER_STATIC_VARIABLE(dbus, dbus_conn) = NULL;
  SERVER_STATIC_VARIABLE(dbus, dbus_list) = NULL;
  SERVER_STATIC_VARIABLE(dbus, dbus_version) = DBUS_VERSION_NONE;

  /* load the library */
  {
    const char dbus_libname0[] = "libdbus-1.so.0";
    const char dbus_libname1[] = "libdbus-1.so.1";
    const char dbus_libname2[] = "libdbus-1.so.2";
    const char dbus_libname3[] = "libdbus-1.so.3";
    const char *libname;
    int version;
    void *lib;
    
    /* DBus 1.0.0 comes with a new library version */
    version = DBUS_VERSION_1_0_0;
    libname = dbus_libname3;
    lib = wine_dlopen(libname, RTLD_NOW, NULL, 0);
    if (!lib)
    {
      version = DBUS_VERSION_0_60_0;
      libname = dbus_libname2;
      lib = wine_dlopen(libname, RTLD_NOW, NULL, 0);
    }
    if (!lib)
    {
      version = DBUS_VERSION_0_35_2;
      libname = dbus_libname1;
      lib = wine_dlopen(libname, RTLD_NOW, NULL, 0);
    }
    if (!lib)
    {
      version = DBUS_VERSION_0_23_4;
      libname = dbus_libname0;
      lib = wine_dlopen(libname, RTLD_NOW, NULL, 0);
      /* verify dbus version */
      if(lib && wine_dlsym(lib, "dbus_connection_has_messages_to_send", NULL, 0) == 0)
      {
        if (debug_level) fprintf(stderr, "dbus version < 0.23; ignoring dbus.\n");
        wine_dlclose(lib, NULL, 0);
        lib = NULL;
      }
    }
    if (!lib)
      return;
    if (debug_level) fprintf(stderr, "using %s\n", libname);
    SERVER_STATIC_VARIABLE(dbus, dbus_lib) = lib;
    SERVER_STATIC_VARIABLE(dbus, dbus_version) = version;
  }
  load_dbus_table();

  /* initialize error handling */
  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_error_init(
      &SERVER_STATIC_VARIABLE(dbus, dbus_error));

  /* open connection */
  SERVER_STATIC_VARIABLE(dbus, dbus_conn) = SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_bus_get(
      DBUS_BUS_SYSTEM, 
      &SERVER_STATIC_VARIABLE(dbus, dbus_error));

  if (!SERVER_STATIC_VARIABLE(dbus, dbus_conn))
  {
    if( debug_level )
      fprintf (stderr, "Couldn't get bus\n");
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_error_free(&SERVER_STATIC_VARIABLE(dbus, dbus_error));
    goto cleanup;
  }

  /* set the watch functions */
  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_set_watch_functions(
      SERVER_STATIC_VARIABLE(dbus, dbus_conn),
      add_dbus_watch, 
      remove_dbus_watch, 
      watch_dbus_toggled, 
      NULL, 
      NULL);

  /* add a match to catch the signals (cdrom notify is a signal) */
  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_bus_add_match(
      SERVER_STATIC_VARIABLE(dbus, dbus_conn),
      "type='signal'",
      &SERVER_STATIC_VARIABLE(dbus, dbus_error));
  if (SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_error_is_set (&SERVER_STATIC_VARIABLE(dbus, dbus_error)))
  {
    if( debug_level )
      fprintf (stderr, "Couldn't add match\n");
    goto cleanup;
  }

  /* add filter for HAL signals */
  if (!SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_add_filter (
       SERVER_STATIC_VARIABLE(dbus, dbus_conn), 
       hal_filter_func, 
       NULL, 
       NULL))
  {
    if( debug_level )
      fprintf (stderr, "Couldn't add filter!\n");
    goto cleanup;
  }

  if (debug_level) fprintf(stderr, "dbus started\n");
  return;


  /* error handling */
cleanup:
  if (debug_level) fprintf(stderr, "dbus not working; shutting down..\n");
  if (SERVER_STATIC_VARIABLE(dbus, dbus_conn))
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_disconnect(SERVER_STATIC_VARIABLE(dbus, dbus_conn));

  wine_dlclose(SERVER_STATIC_VARIABLE(dbus, dbus_lib), NULL, 0);
  SERVER_STATIC_VARIABLE(dbus, dbus_lib) = NULL;
}


void dbus_release(void)
{
  if (debug_level) fprintf(stderr, "shutting down dbus..\n");

  /* if dbus didn't init, don't do anything */
  if (!SERVER_STATIC_VARIABLE(dbus, dbus_lib))
    return;

  /* remove filter */
  SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_remove_filter (
       SERVER_STATIC_VARIABLE(dbus, dbus_conn), 
       hal_filter_func, 
       NULL);

  /* close the connection */
  if (SERVER_STATIC_VARIABLE(dbus, dbus_conn))
    SERVER_STATIC_VARIABLE(dbus, dbus_ltable).dbus_connection_disconnect(SERVER_STATIC_VARIABLE(dbus, dbus_conn));

  /* close the library */
  wine_dlclose(SERVER_STATIC_VARIABLE(dbus, dbus_lib), NULL, 0);
}

void dbus_stop(void)
{
    /* Don't need to do anything here, but we might later potentially */
}
