/*
 * Server-side support for MacOS X
 *
 * Copyright (C) 2006 TransGaming Technologies
 */

#include "config.h"
#include "wine/port.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>

#include "process.h"
#include "thread.h"
#include "request.h"
#include "server_memory.h"


/* handle a SIGCHLD signal */
void sigchld_handler( int sig )
{
}


/* detach from a Unix thread and kill it */
void detach_thread( struct thread *thread, int sig )
{
    if (!thread->unix_tid_or_pid) return;
    if (sig)
    {
       thread_terminate (thread->thread_port);
       return;
    }

    if (thread->suspend + thread->process->suspend > 0)
       continue_thread (thread);
}

/* stop a thread (at the Unix level) */
void stop_thread( struct thread *thread )
{
   mach_msg_return_t Err;

   /* can't stop a thread while initialisation is in progress */
   if (!thread->unix_tid_or_pid || thread->process->init_event ||
       !thread->thread_port)
      return;

   Err = thread_suspend (thread->thread_port);
   if (Err != MACH_MSG_SUCCESS)
      fprintf (stderr, "Warn: suspending thread port 0x%x failed: 0x%x\n",
               thread->thread_port, Err);
   else
   {
      Err = thread_abort (thread->thread_port);
      if ((Err != MACH_MSG_SUCCESS) && debug_level)
         fprintf (stderr, "Warn: aborting thread port 0x%x failed: 0x%x\n",
                  thread->thread_port, Err);
   }
}

/* make a thread continue (at the Unix level) */
void continue_thread( struct thread *thread )
{
   mach_msg_return_t Err;

   if (!thread->unix_tid_or_pid || !thread->thread_port)
      return;

   Err = thread_resume (thread->thread_port);
   if (Err != MACH_MSG_SUCCESS)
      fprintf (stderr, "Warn: resuming thread port 0x%x failed: 0x%x\n",
               thread->thread_port, Err);
}

/* suspend thread for interaction */
int suspend_for_ptrace( struct thread *thread )
{
   suspend_thread (thread, 0);
   return 1;
}


/* read data from specified process */
int read_thread_data (struct thread *thread, uintptr_t addr, size_t len,
		      void* data, size_t* p_num_read)
{
    kern_return_t Err;
    vm_size_t bytes_read = len;

    Err = vm_read_overwrite (thread->process->task_port, addr, len,
                             (vm_address_t)data,
                             &bytes_read);

    if (Err != KERN_SUCCESS)
       return -1;

    if (p_num_read)
       *p_num_read = bytes_read;

    return 0;
}


/* read data from a process memory space
 * addr is the virtual address in that process, len is the number of bytes.
 */
void read_process_memory (struct process *process, uintptr_t addr,
                          size_t len, void *dest, size_t check_len)
{
    struct thread *thread = process->thread_list;
    vm_size_t bytes_read = len;
    kern_return_t Err;

    if (!thread)  /* process is dead */
    {
        set_error (STATUS_ACCESS_DENIED);
        return;
    }

    set_error (STATUS_SUCCESS);
    if (!(thread->sched_mode & SCDM_INTERNAL))
       suspend_for_ptrace (thread);

    Err = vm_read_overwrite (thread->process->task_port, addr, len,
                             (vm_address_t)dest,
                             &bytes_read);

    if (!(thread->sched_mode & SCDM_INTERNAL))
       resume_thread (thread);

    if (Err == KERN_PROTECTION_FAILURE)
       set_error (STATUS_ACCESS_DENIED);
    else if (Err != KERN_SUCCESS)
       set_error (STATUS_ACCESS_VIOLATION);

    if (Err != KERN_SUCCESS)
       fprintf (stderr, "Warn: read_process_memory failed: 0x%x\n", Err);
}


/* write data to a process memory space
 * addr is the virtual address in that process, len is the number of bytes.
 */
void write_process_memory (struct process *process, uintptr_t addr,
                           size_t len, const void *src,
                           size_t check_len)
{
    struct thread *thread = process->thread_list;
    vm_address_t RegionAddr = addr;
    vm_size_t RegionSize;
    struct vm_region_basic_info Info;
    mach_msg_type_number_t InfoSize;
    mach_port_t ObjectName;
    kern_return_t Err;

    if (!thread)  /* process is dead */
    {
        set_error( STATUS_ACCESS_DENIED );
        return;
    }

    set_error (STATUS_SUCCESS);
    if (!(thread->sched_mode & SCDM_INTERNAL))
       suspend_for_ptrace (thread);

    InfoSize = sizeof (Info);
    if ((Err = vm_region (thread->process->task_port, &RegionAddr,
                          &RegionSize, VM_REGION_BASIC_INFO,
                          (vm_region_info_t)&Info,
                          &InfoSize, &ObjectName)))
    {
       fprintf (stderr, "Warn: vm_region failed: 0x%x\n", Err);
       set_error (STATUS_ACCESS_DENIED);
       resume_thread (thread);
       return;
    }

    if (!(Info.protection & VM_PROT_WRITE))
    {
       if ((Err = vm_protect (thread->process->task_port, addr, len, 0,
                              Info.protection | VM_PROT_WRITE)))
       {
          fprintf (stderr, "Warn: vm_protect failed: 0x%x\n", Err);
          set_error (STATUS_ACCESS_DENIED);
          resume_thread (thread);
          return;
       }
    }

    Err = vm_write (thread->process->task_port, addr, (vm_offset_t)src, len);
    if (Err == KERN_PROTECTION_FAILURE)
       set_error (STATUS_ACCESS_DENIED);
    else if (Err != KERN_SUCCESS)
       set_error (STATUS_ACCESS_VIOLATION);

    if (Err != KERN_SUCCESS)
       fprintf (stderr, "Warn: write_process_memory failed: 0x%x\n", Err);

    if (!(Info.protection & VM_PROT_WRITE))
    {
       if ((Err = vm_protect (thread->process->task_port, addr, len, 0,
                              Info.protection)))
       {
          fprintf (stderr, "Warn: second vm_protect failed: 0x%x\n", Err);
          set_error (STATUS_ACCESS_DENIED);
          resume_thread (thread);
          return;
       }
    }

    resume_thread (thread);
}
