/* Stuff for scheduling threads manually
 *
 * Copyright 2004-2005 TransGaming Technologies
 *
 * Remaining Issues
 *  - new thread should run before old thread to fix MSI issue,
 *    foreground threads should get higher priority,
 *    threads who just received a message should get temporary boosts,
 *    and so on.
 *  - scheduler: timer may not switch out the best task
 *  - timer is real-time-based, not CPU-time-based, so linux tasks may
 *    make scheduled threads not get a chance to run
 *  - only for single-CPU machines right now
 *
 * Test cases:
 *  Vice City may depend on starved low-priority threads being allowed to run
 *  now and then; some scenery detail will not show up otherwise.
 *  Alternatively, it may depend on the 1-point priority boost given to
 *  threads who wait for an event and get one. (This boost does seem to make
 *  scenery load much faster than just regular anti-starvation does.)
 */

#include <stdio.h>
#include <unistd.h>

#include "request.h"
#include "thread.h"
#include "process.h"
#include "server_memory.h"

#define PRIOS 32
#define RTPRIO 16

/* NT server use 120ms slices;
 * NT Workstation can be configured with 20ms, 40ms, or 60ms slices for foreground threads;
 * background threads get 20ms slices.
 * They normally measure the slices with a resolution of 10ms. */
#define TIME_RESOLUTION 30
#define TIME_SLICE_COUNT 2 /* 2*30ms = 60ms slices */

/* NT gives a priority boost if a thread has been starving for 3 seconds.
 * (This boost maxes the priority to 15 for a single time slice, but
 * this slice is double as long as a normal slice.) */
#define MAX_STARVE 3000 /* 3 seconds */

START_FILE_STATIC_VARIABLES(schedule)
  int use_scheduler;
  struct thread *sched_queue[PRIOS];
  struct thread *sched_current;
  struct thread *sched_wait_queue;
  struct timeout_user *sched_timer;
  struct timeval sched_last_shuffle;
  struct timeval sched_next_shuffle;
END_FILE_STATIC_VARIABLES(schedule)

static int is_foreground(struct thread* thread)
{
    /* always assume foreground for now. */
    return 1;
}

static int basic_priority(int base, int pri)
{
    if (pri <= THREAD_PRIORITY_IDLE) return 1;
    if (pri >= THREAD_PRIORITY_TIME_CRITICAL) return 15;
    if (pri < THREAD_PRIORITY_LOWEST) return base-2;
    if (pri > THREAD_PRIORITY_HIGHEST) return base+2;
    return base + pri;
}

static int rt_priority(int pri)
{
    if (pri <= THREAD_PRIORITY_IDLE) return 16;
    if (pri >= THREAD_PRIORITY_TIME_CRITICAL) return 31;
    return 24 + pri;
}

static int get_base_priority(struct thread* thread)
{
    int priority_class = thread->process->priority;

    if (thread->sched_mode & SCDM_INTERNAL)
        priority_class = REALTIME_PRIORITY_CLASS;

    switch (priority_class) {
    case IDLE_PRIORITY_CLASS:
        return basic_priority(4, thread->priority);
    case NORMAL_PRIORITY_CLASS:
    default:
        return basic_priority(is_foreground(thread) ? 9 : 7, thread->priority);
    case HIGH_PRIORITY_CLASS:
        return basic_priority(13, thread->priority);
    case REALTIME_PRIORITY_CLASS:
        return rt_priority(thread->priority);
#if 0 /* not defined in winbase.h */
    case BELOW_NORMAL_PRIORITY_CLASS:
        return basic_priority(6, thread->priority);
    case ABOVE_NORMAL_PRIORITY_CLASS:
        return basic_priority(10, thread->priority);
#endif
    }
}

static int get_sched_priority(struct thread* thread)
{
    int prio = get_base_priority(thread);
    if (prio >= 16) return prio;
    if (thread->sched_boost < 0) return 15;
    prio += thread->sched_boost;
    if (prio >= 16) return 15;
    return prio;
}

void check_wait_queue(void)
{
extern int is_wakeup;
    struct thread *next;

    /* Scan all waiting threads to see whether any waits have been satisfied,
     * and if so, wake them from here. Normally they will wake themselves up,
     * but sometimes it takes far too long before they get that opportunity.
     * Doing it from here will ensure that it doesn't take too long before
     * it wakes up. */
    next = SERVER_STATIC_VARIABLE(schedule,sched_wait_queue);
    is_wakeup = 1;
    while (next) {
        wake_thread(next);
        next = next->sched_next;
    }
    is_wakeup = 0;
}

static void shuffle_threads(void* priv)
{
    struct thread *thread;
    struct thread *next;
    struct timeval last_shuffle;
    struct timeval next_shuffle;
    int queue;

    thread = SERVER_STATIC_VARIABLE(schedule,sched_current);
    last_shuffle = SERVER_STATIC_VARIABLE(schedule,sched_last_shuffle);
    next_shuffle = SERVER_STATIC_VARIABLE(schedule,sched_next_shuffle);

    if (thread) {
        info_msg( WITH_DEBUG, thread, "--tick--\n" );

#if 0
        /* shuffle all threads that executed in the last timeslice;
         * that way the most starved thread is always first in line
         * in its queue. */
        for (queue=0; queue<PRIOS; queue++) {
            next = SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue];
            if (!next) continue;
            if (next->sched_next &&
                (next == thread ||
                 !time_before(&next->sched_last, &last_shuffle))) {
                /* remove from front of queue */
                dequeue_thread(next);
                /* insert into back of queue */
                queue_thread(next);
            }
        }
#endif

        if (thread->sched_left)
            thread->sched_left--;
        if (!thread->sched_left) {
            /* Time slice expired.
             * Decrease any priority boosts and
             * move thread to end of queue. */
            int boost = thread->sched_boost;
            if (boost < 0)
                boost = -(boost + 1);
            else if (boost > 0)
                boost--;
            info_msg( WITH_DEBUG, thread, "-expired- (new boost %d)\n", boost);
            dequeue_thread(thread);
            thread->sched_boost = boost;
            queue_thread(thread);
        }

        /* check_wait_queue(); */

        /* Scan for starving threads. In theory, the most starved thread
         * should normally be first in each queue, so we don't have to
         * check all threads. */
        for (queue=0; queue<RTPRIO; queue++) {
            next = SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue];
            if (next == thread) continue;
            if (next && time_before(&next->sched_before, &next_shuffle) &&
                (next->sched_boost >= 0)) {
                next->sched_boost = -(next->sched_boost + 1);
                next->sched_left = 2 * TIME_SLICE_COUNT;
                info_msg( WITH_DEBUG, next, "-starved- (new boost %d)\n", next->sched_boost);
                requeue_thread(next);
            }
        }

        reschedule();
    }

    /* schedule new timeout */
    SERVER_STATIC_VARIABLE(schedule,sched_last_shuffle) = next_shuffle;
    add_timeout(&next_shuffle, TIME_RESOLUTION);
    SERVER_STATIC_VARIABLE(schedule,sched_next_shuffle) = next_shuffle;
    SERVER_STATIC_VARIABLE(schedule,sched_timer) =
        add_timeout_user(&next_shuffle, shuffle_threads_callback_tag, NULL);
}
timeout_callback get_shuffle_threads_callback(void) { return shuffle_threads; }

static void install_timer(void)
{
    struct timeval next_shuffle;

    if (SERVER_STATIC_VARIABLE(schedule,sched_timer) != NULL)
        return;
    info_msg( WITH_DEBUG, NULL, "--installtimer--\n" );
    gettimeofday(&next_shuffle, 0);
    add_timeout(&next_shuffle, TIME_RESOLUTION);
    SERVER_STATIC_VARIABLE(schedule,sched_next_shuffle) = next_shuffle;
    SERVER_STATIC_VARIABLE(schedule,sched_timer) =
        add_timeout_user(&next_shuffle, shuffle_threads_callback_tag, NULL);
}

void thread_not_starved(struct thread* thread)
{
    struct timeval now;
    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return;
    gettimeofday(&now, 0);
    thread->sched_last = now;
    thread->sched_before = now;
    add_timeout(&thread->sched_before, MAX_STARVE);
}

static void unsleep_thread(struct thread* thread)
{
    struct timeval now = {0,0};
    if (thread->sched_sleep) {
        if (debug_level)
            gettimeofday(&now, 0);
        if (!is_wineserver && current == thread) {
            /* Descheduled thread is the one running the shmserver.
             * We are obviously not sleeping yet, so just cancel the sleep. */
            info_msg( WITH_DEBUG, thread, "-sleep cancelled- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
            thread->sched_sleep--;
        }
        else {
            info_msg( WITH_DEBUG, thread, "-rescheduled from sleep- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
            thread->sched_sleep--;
            if (thread->sched_sleep_cookie) {
                struct wake_up_reply reply;
                int ret;
                reply.cookie = thread->sched_sleep_cookie;
                reply.signaled = 0;
                ret = write( get_context_fd( thread->wait_fd ), &reply, sizeof(reply) );
                thread->sched_sleep_cookie = 0;
            }
        }
    }
}

static void deschedule_thread(struct thread* thread)
{
    struct timeval now = {0,0};
    if (!thread->wait && !thread->sched_sleep && !(thread->sched_mode & SCDM_BLOCKED)) {
        if (debug_level)
            gettimeofday(&now, 0);
        if (!is_wineserver && current == thread) {
            /* Descheduled thread is the one running the shmserver.
             * Stop it by having it do a blocking read in shared_server_call. */
            info_msg( WITH_DEBUG, thread, "-descheduled with sleep- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
            thread->sched_sleep++;
        }
        else if (thread->process->init_event) {
            /* Descheduled thread is currently booting.
             * Sending a signal is too dangerous. */
            info_msg( WITH_DEBUG, thread, "-descheduled during init- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
            thread->sched_sleep++;
        } else {
            info_msg( WITH_DEBUG, thread, "-descheduled with signal- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
            thread->sched_sleep++;
            send_unix_signal(thread, 0, 0, 0, NULL);
        }
    }
}

static void reschedule_thread(struct thread* thread)
{
    if (!thread->sched_left) {
        thread->sched_left = TIME_SLICE_COUNT;
    }
    if (thread->sched_sleep) {
        unsleep_thread(thread);
    }
}

static void preschedule_thread(struct thread* thread)
{
    if (thread->sched_sleep) {
#if 0
        /* replaces the sleep with a suspend, so that
         * the wait_fd can be used for other things. */
        thread->sched_sleep--;
        deschedule_thread(thread);
        thread->sched_sleep++;
        unsleep_thread(thread);
#endif
    }
    else {
        /* suspend thread if we're about to send it a wakeup,
         * but don't want it to actually run yet. */
        deschedule_thread(thread);
    }
}

/* boost priority of thread */
int boost_thread(struct thread* thread, int boost)
{
    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return thread->sched_queue;

    if (thread->sched_boost < 0)
        thread->sched_boost -= boost;
    else
        thread->sched_boost += boost;
    info_msg( WITH_DEBUG, thread, "-boosted- (new boost %d)\n", thread->sched_boost);
    return requeue_thread(thread);
}

/* remove thread from wait queue. */
void dequeue_thread_wait(struct thread* thread)
{
    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return;

    if (thread->sched_queue != -1)
        return;

    info_msg( WITH_DEBUG, thread, "--dequeue wait--\n" );

    if (thread->sched_next != NULL) {
        thread->sched_next->sched_prev = thread->sched_prev;
    }
    if (thread->sched_prev != NULL) {
        thread->sched_prev->sched_next = thread->sched_next;
    } else {
        if (SERVER_STATIC_VARIABLE(schedule,sched_wait_queue) == thread)
            SERVER_STATIC_VARIABLE(schedule,sched_wait_queue) = thread->sched_next;
    }
    thread->sched_prev = NULL;
    thread->sched_next = NULL;
}

/* insert thread into wait queue. */
void queue_thread_wait(struct thread* thread)
{
    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return;

    if (thread->sched_queue != -1)
        return;

    if (thread->sched_next || thread->sched_prev ||
        SERVER_STATIC_VARIABLE(schedule,sched_wait_queue) == thread)
        return;

    info_msg( WITH_DEBUG, thread, "--queue wait--\n" );

    thread->sched_next = SERVER_STATIC_VARIABLE(schedule,sched_wait_queue);
    if (thread->sched_next != NULL)
        thread->sched_next->sched_prev = thread;
    SERVER_STATIC_VARIABLE(schedule,sched_wait_queue) = thread;
}

/* remove thread from scheduling queue. */
int dequeue_thread(struct thread* thread)
{
    int queue;

    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return thread->sched_queue;

    info_msg( WITH_DEBUG, thread, "--dequeue--\n" );
    queue = thread->sched_queue;
    if (queue == -1)
        return queue;
    if (thread->sched_next != NULL) {
        thread->sched_next->sched_prev = thread->sched_prev;
    }
    if (thread->sched_prev != NULL) {
        thread->sched_prev->sched_next = thread->sched_next;
    } else {
        if (SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue] == thread)
            SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue] = thread->sched_next;
    }
    thread->sched_prev = NULL;
    thread->sched_next = NULL;
    thread->sched_queue = -1;
    /* not sure if remainder of timeslice should be abandoned,
     * nor the current priority boosts. I'll assume they should.
     */
    thread->sched_left = 0;
    thread->sched_boost = 0;

    /* if a signal is pending, and this dequeue call is because the
     * client just decided to start waiting for some other reason,
     * we should cancel the pending signal, or we may deadlock the wait. */
    if (thread->wait && thread->sched_sleep && !thread->sched_sleep_cookie) {
        thread->sched_sleep--;
    }

    info_msg( WITH_DEBUG, thread, "-dequeued- (queue %d)\n", queue);
    return queue;
}

/* insert thread into scheduling queue. */
int queue_thread(struct thread* thread)
{
    int queue, run_now;
    struct thread *prev;
    struct timeval now = {0,0};

    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler)) {
        /* recv_unix_signal may invoke sched_sleep even when scheduler is off,
         * handle this situation here. */
        if (thread->sched_sleep)
            unsleep_thread(thread);
        return thread->sched_queue;
    }

    if (thread->sched_mode & SCDM_BLOCKED) {
        /* shouldn't happen */
        info_msg( ALWAYS, thread, "BUG: attempting to queue a blocked thread\n" );
        return -1;
    }

    if (debug_level)
        gettimeofday(&now, 0);
    info_msg( WITH_DEBUG, thread, "--queue-- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
    if (thread->sched_queue != -1)
        return thread->sched_queue;
    prev = SERVER_STATIC_VARIABLE(schedule,sched_current);
    queue = get_sched_priority(thread);
    if (SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue] == NULL) {
        SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue] = thread;
        info_msg( WITH_DEBUG, thread, "queued first\n" );
        run_now = 1;
    } else {
        struct thread* other;
        other = SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue];
        if (thread->sched_left && other != prev) {
            /* time left in its slice - queue first */
            thread->sched_next = other;
            other->sched_prev = thread;
            SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue] = thread;
            info_msg( WITH_DEBUG, thread, "queued first\n" );
            run_now = 1;
        } else {
            if (!thread->sched_left) {
                /* no time left in its slice - queue last */
                while (other->sched_next != NULL)
                    other = other->sched_next;
            }
            /* else queue after currently running thread */
            thread->sched_prev = other;
            if (other->sched_next)
                other->sched_next->sched_prev = thread;
            thread->sched_next = other->sched_next;
            other->sched_next = thread;
            info_msg( WITH_DEBUG, thread, "queued after %04x\n", get_thread_id(other) );
            run_now = 0;
        }
    }
    thread->sched_queue = queue;
    /* if the thread was just woken from a wait,
     * suspend it until CPU time is allocated to it. */
    info_msg( WITH_DEBUG, thread, "wait=%p, slp=%d, susp=%d\n", thread->wait,
              thread->sched_sleep, thread->process->suspend + thread->suspend );
    if (run_now && prev && (prev != thread) &&
        (prev->sched_queue >= queue))
        run_now = 0;
    if (run_now) {
        /* This thread is high priority - schedule it for immediate execution.
         * More efficient than first prescheduling then rescheduling
         * (and obviates the need for an immediate reschedule). */
        if (prev && (prev != thread)) {
            deschedule_thread(prev);
        }
        SERVER_STATIC_VARIABLE(schedule,sched_current) = thread;
        if (prev && (prev != thread))
            thread_not_starved(prev);
        reschedule_thread(thread);
    }
    else {
        preschedule_thread(thread);
    }
    if (debug_level)
        gettimeofday(&now, 0);
    info_msg( WITH_DEBUG, thread, "-queued- (queue %d) (at %ld:%ld)\n", queue, now.tv_sec, now.tv_usec );
    return queue;
}

/* move thread between scheduling queues. */
int requeue_thread(struct thread* thread)
{
    int queue = get_sched_priority(thread);
    int boost = thread->sched_boost;
    int left = thread->sched_left;
    /* only requeue if already queued. */
    if (thread->sched_queue == -1)
        return queue;
    /* only requeue if priority changed. */
    if (thread->sched_queue == queue)
        return queue;
    dequeue_thread(thread);
    thread->sched_boost = boost;
    thread->sched_left = left;
    queue_thread(thread);
    return queue;
}

void requeue_process(struct process* process)
{
    struct thread *thread;
    for (thread = process->thread_list; thread; thread = thread->proc_next)
    {
        requeue_thread(thread);
    }
}

void yield_thread(struct thread* thread, int extra_boost)
{
    int boost = thread->sched_boost;
    int left = thread->sched_left;
    /* only requeue if already queued. */
    if (thread->sched_queue == -1)
        return;
    /* if there's nobody else in the queue, don't bother. */
    if (thread->sched_next == NULL)
        return;
    dequeue_thread(thread);
    /* I'm assuming we should not drop the boost...
     * but perhaps we should at least decay it? */
    thread->sched_boost = boost;
    /* apply additional boost now to avoid redundant requeues */
    if (extra_boost) boost_thread(thread, extra_boost);
    check_wait_queue();
    queue_thread(thread);
    /* give back remainder of original timeslice, so that
     * priority boosts still decay in finite time */
    thread->sched_left = left;
}

/* execute the highest priority thread. */
void reschedule(void)
{
    struct thread *next;
    struct thread *prev;
    int queue;
    struct timeval now = {0,0};

    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return;

    if (debug_level)
        gettimeofday(&now, 0);
    info_msg( WITH_DEBUG, is_wineserver ? NULL : current,
                          "--reschedule-- (at %ld:%ld)\n", now.tv_sec, now.tv_usec );
    /* schedule highest priority queued thread. */
    next = NULL;
    for (queue=PRIOS; queue>0; queue--) {
        next = SERVER_STATIC_VARIABLE(schedule,sched_queue)[queue-1];
        if (next) break;
    }
    prev = SERVER_STATIC_VARIABLE(schedule,sched_current);
    if (prev && prev != next) {
        if (debug_level)
            gettimeofday(&now, 0);
        info_msg( WITH_DEBUG, prev, "-descheduled- (queue %d) (at %ld:%ld)\n", prev->sched_queue, now.tv_sec, now.tv_usec );
        deschedule_thread(prev);
        thread_not_starved(prev);
    }
    SERVER_STATIC_VARIABLE(schedule,sched_current) = next;
    if (next) {
        reschedule_thread(next);
        if (debug_level)
            gettimeofday(&now, 0);
        info_msg( WITH_DEBUG, next, "-scheduled- (queue %d) (at %ld:%ld)\n", next->sched_queue, now.tv_sec, now.tv_usec );
    }
}

void initschedule(struct thread* thread, int suspended)
{
    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return;
    info_msg( WITH_DEBUG, thread, "--initschedule--\n" );
    if (suspended <= 0) {
        thread_not_starved(thread);
        queue_thread(thread);
        reschedule();
    }
    install_timer();
}

void deinitschedule(struct thread* thread)
{
    int queue;

    if (!SERVER_STATIC_VARIABLE(schedule,use_scheduler))
        return;
    info_msg( WITH_DEBUG, thread, "--deinitschedule--\n" );
    queue = dequeue_thread(thread);
    dequeue_thread_wait(thread);
    if (SERVER_STATIC_VARIABLE(schedule,sched_current) == thread) {
        SERVER_STATIC_VARIABLE(schedule,sched_current) = NULL;
    }
    if (queue != -1)
        check_wait_queue();
    reschedule();
}

void select_scheduler(int use_scheduler)
{
    SERVER_STATIC_VARIABLE(schedule,use_scheduler) = use_scheduler;
}

int set_scheduling_mode(struct thread* thread, int mode)
{
    int old_mode = thread->sched_mode;
    int old_queue = thread->sched_queue;

    /* don't drop the internal flag */
    mode |= (old_mode & (SCDM_INTERNAL | SCDM_CREATED_SUSPENDED));

    if ((mode & SCDM_INTERNAL) && !(old_mode & SCDM_INTERNAL) && (thread->suspend > 0))
        mode |= SCDM_CREATED_SUSPENDED;

    if (mode == old_mode)
        return old_mode;

    if (!(thread->sched_mode & SCDM_BLOCKED)) {
        dequeue_thread(thread);
        /* don't check_wait_queue() here, this isn't a "regular" dequeue, and
         * since it happens fairly often during I/O, the queue-checking overhead
         * would generally not be worth it in such circumstances. */
    }
    thread->sched_mode = mode;
    if (!(thread->sched_mode & SCDM_BLOCKED)) {
        if (old_mode & SCDM_BLOCKED)
            thread_not_starved(thread);
        if ((old_mode & SCDM_BLOCKED) || old_queue != -1)
            queue_thread(thread);
    }
    reschedule();

    /* If there's a pending deschedule on a thread that is entering
     * client-side blocking, cancel the signal, otherwise the client
     * will never run again. */
    if ((thread->sched_mode & SCDM_BLOCKED) && thread->sched_sleep && !thread->sched_sleep_cookie) {
        thread->sched_sleep--;
    }

    return old_mode;
}
