/*
 * Server-side semaphore management
 *
 * Copyright (C) 1998 Alexandre Julliard
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "winnt.h"

#include "handle.h"
#include "thread.h"
#include "request.h"

struct semaphore
{
    struct object  obj;    /* object header */
    unsigned int   count;  /* current count */
    unsigned int   max;    /* maximum possible count */
};

static void semaphore_dump( struct object *obj, int verbose );
static int semaphore_signaled( struct object *obj, struct wait_queue_entry *entry );
static int semaphore_do_signal (struct object *obj, unsigned int access);
static int semaphore_satisfied( struct object *obj, struct thread *thread );

static const struct object_ops semaphore_ops =
{
    sizeof(struct semaphore),      /* size */
    semaphore_dump,                /* dump */
    add_queue,                     /* add_queue */
    remove_queue,                  /* remove_queue */
    semaphore_signaled,            /* signaled */
    semaphore_do_signal,           /* signal */
    semaphore_satisfied,           /* satisfied */
    NULL,                          /* get_poll_events */
    NULL,                          /* poll_event */
    no_get_fd,                     /* get_fd */
    no_get_ctx_fd,                 /* get_ctx_fd */
    no_flush,                      /* flush */
    no_get_file_info,              /* get_file_info */
    NULL,                          /* queue_async */
    no_destroy                     /* destroy */
};
const struct object_ops* get_this_process_semaphore_ops(void) { return &semaphore_ops; }

static struct semaphore *create_semaphore( const WCHAR *name, size_t len,
                                           unsigned int initial, unsigned int max )
{
    struct semaphore *sem;

    if (!max || (initial > max))
    {
        set_error( STATUS_INVALID_PARAMETER );
        return NULL;
    }
    if ((sem = create_named_object( semaphore_ops_tag, name, len )))
    {
        if (get_error() != STATUS_OBJECT_NAME_COLLISION)
        {
            /* initialize it if it didn't already exist */
            sem->count = initial;
            sem->max   = max;
        }
    }
    return sem;
}

static int release_semaphore (struct semaphore *sem, unsigned int count,
                              unsigned int *pPrev)
{
   if (pPrev)
      *pPrev = sem->count;

   if ((sem->count + count < sem->count) || (sem->count + count > sem->max))
   {
      set_error (STATUS_SEMAPHORE_LIMIT_EXCEEDED);
      return 0;
   }

   if (sem->count)
   {
      /* there cannot be any thread waiting if the count is != 0 */
      assert (!sem->obj.head);
      sem->count += count;
   }
   else
   {
      sem->count = count;
      wake_up (&sem->obj, count);
   }

   return 1;
}


static void semaphore_dump( struct object *obj, int verbose )
{
    struct semaphore *sem = (struct semaphore *)obj;
    assert( obj->ops == semaphore_ops_tag );
    fprintf( stderr, "Semaphore count=%d max=%d ", sem->count, sem->max );
    dump_object_name( &sem->obj );
    fputc( '\n', stderr );
}

static int semaphore_signaled( struct object *obj, struct wait_queue_entry *entrys )
{
    struct semaphore *sem = (struct semaphore *)obj;
    assert( obj->ops == semaphore_ops_tag );
    return (sem->count > 0);
}

/* Not named semaphore_signal() as that conflicts with a function
   on Apple */
static int semaphore_do_signal (struct object *obj, unsigned int access)
{
   struct semaphore *sem = (struct semaphore *)obj;

   assert (obj->ops == semaphore_ops_tag);

   if (!(access & SEMAPHORE_MODIFY_STATE))
   {
      set_error (STATUS_ACCESS_DENIED);
      return 0;
   }

   return release_semaphore (sem, 1, NULL);
}

static int semaphore_satisfied( struct object *obj, struct thread *thread )
{
    struct semaphore *sem = (struct semaphore *)obj;
    assert( obj->ops == semaphore_ops_tag );
    assert( sem->count );
    sem->count--;
    return 0;  /* not abandoned */
}

/* create a semaphore */
DECL_HANDLER(create_semaphore)
{
    struct semaphore *sem;

    reply->handle = 0;
    if ((sem = create_semaphore( get_req_data(), get_req_data_size(),
                                 req->initial, req->max )))
    {
        reply->handle = alloc_handle( current->process, sem, SEMAPHORE_ALL_ACCESS, req->inherit );
        release_object( sem );
    }
}

/* open a handle to a semaphore */
DECL_HANDLER(open_semaphore)
{
    reply->handle = open_object( get_req_data(), get_req_data_size(),
                                 semaphore_ops_tag, req->access, req->inherit );
}

/* release a semaphore */
DECL_HANDLER(release_semaphore)
{
   struct semaphore *sem;

   if ((sem = (struct semaphore *)get_handle_obj (current->process,
                                                  req->handle,
                                                  SEMAPHORE_MODIFY_STATE,
                                                  semaphore_ops_tag)))
   {
      release_semaphore (sem, req->count, &reply->prev_count);
      release_object (sem);
   }
}
